class Operator:

    def __init__(self, operator_label : str = '', function = None, min_arity : int = -1, max_arity : int = -1):
        self.operator_label = operator_label
        self.function = function
        self.max_arity = max_arity
        self.min_arity = min_arity

    def exec(self, args):
        return self.function(args)

class OperatorReturn(Operator):

    def __init__(self, operator_label: str = 'return', function = None):
        super().__init__(operator_label, function, min_arity=1, max_arity=1)

    def exec(self, args):
        if len(args) > self.max_arity or len(args) < self.min_arity:
            raise AssertionError(
                f"Number of arguments do not match the allowed arity min:{self.min_arity} max:{self.max_arity}")
        return args[0]

class OperatorAdd(Operator):

    def __init__(self, operator_label: str = 'add', function = None):
        super().__init__(operator_label, function)

    def exec(self, args):
        a = 0
        for v in args:
            a += v
        return a

class OperatorSub(Operator):

    def __init__(self, operator_label: str = 'sub', function = None):
        super().__init__(operator_label, function)

    def exec(self, args):
        a = args[0]
        for v in args[1:]:
            a -= v
        return a

class OperatorMul(Operator):

    def __init__(self, operator_label: str = 'mul', function = None):
        super().__init__(operator_label, function)

    def exec(self, args):
        a = args[0]
        for v in args[1:]:
            a = a * v
        return a

class OperatorDiv(Operator):

    def __init__(self, operator_label: str = 'div', function = None):
        super().__init__(operator_label, function)

    def exec(self, args):
        a = args[0]
        for v in args[1:]:
            a = a / v
        return a

class OperatorPow(Operator):

    def __init__(self, operator_label: str = 'pow', function = None):
        super().__init__(operator_label, function)

    def exec(self, args):
        a = args[0]
        for v in args[1:]:
            a = a ** v
        return a

class OperatorLTE(Operator):

    def __init__(self, operator_label: str = 'lte', function = None):
        super().__init__(operator_label, function, min_arity=2, max_arity=2)

    def exec(self, args):
        if len(args) > self.max_arity or len(args) < self.min_arity:
            raise AssertionError(
                f"Number of arguments do not match the allowed arity min:{self.min_arity} max:{self.max_arity}")

        return args[0] <= args[1]

class OperatorLT(Operator):

    def __init__(self, operator_label: str = 'lt', function=None):
        super().__init__(operator_label, function, min_arity=2, max_arity=2)

    def exec(self, args):
        if len(args) > self.max_arity or len(args) < self.min_arity:
            raise AssertionError(
                f"Number of arguments do not match the allowed arity min:{self.min_arity} max:{self.max_arity}")

        return args[0] < args[1]

class OperatorGTE(Operator):

    def __init__(self, operator_label: str = 'gte', function=None):
        super().__init__(operator_label, function, min_arity=2, max_arity=2)

    def exec(self, args):
        if len(args) > self.max_arity or len(args) < self.min_arity:
            raise AssertionError(
                f"Number of arguments do not match the allowed arity min:{self.min_arity} max:{self.max_arity}")

        return args[0] >= args[1]

class OperatorGT(Operator):

    def __init__(self, operator_label: str = 'gt', function=None):
        super().__init__(operator_label, function, min_arity=2, max_arity=2)

    def exec(self, args):
        if len(args) > self.max_arity or len(args) < self.min_arity:
            raise AssertionError(
                f"Number of arguments do not match the allowed arity min:{self.min_arity} max:{self.max_arity}")

        return args[0] > args[1]

class OperatorEq(Operator):

    def __init__(self, operator_label: str = 'eq', function=None):
        super().__init__(operator_label, function, min_arity=2, max_arity=2)

    def exec(self, args):
        if len(args) > self.max_arity or len(args) < self.min_arity:
            raise AssertionError(
                f"Number of arguments do not match the allowed arity min:{self.min_arity} max:{self.max_arity}")

        return args[0] == args[1]

class OperatorNE(Operator):

    def __init__(self, operator_label: str = 'ne', function=None):
        super().__init__(operator_label, function, min_arity=2, max_arity=2)

    def exec(self, args):
        if len(args) > self.max_arity or len(args) < self.min_arity:
            raise AssertionError(
                f"Number of arguments do not match the allowed arity min:{self.min_arity} max:{self.max_arity}")

        return args[0] != args[1]


import unittest
from natural_selection.genetic_algorithms import Gene, Chromosome, Individual, Island
from natural_selection.genetic_algorithms.utils.random_functions import random_int
from natural_selection.genetic_algorithms.operators.crossover import crossover_two_n_point
class TestSimpleIsland(unittest.TestCase):

    def setUp(self) -> None:
        self.fitness = lambda chromosome, island, x, y: (chromosome[0].value * x) + (chromosome[1].value *y)
        g_1 = Gene(name="first", value=1, gene_max=25, gene_min=1, randomise_function=random_int)
        g_2 = Gene(name="second", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_3 = Gene(name="third", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_4 = Gene(name="fourth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_5 = Gene(name="fifth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_6 = Gene(name="sixth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        gen = Chromosome([g_1, g_2, g_3, g_4, g_5, g_6])
        self.ind = Individual(self.fitness, name="Adam", chromosome=gen)

        self.life = Island({'x': 0.6, 'y' : 0.2}, crossover_function=crossover_two_n_point)

    def test_create(self):
        self.life.initialise(self.ind, population_size=5)
        self.assertEquals(len(self.life.population), 5)

    def test_import_migrants(self):
        self.life.initialise(self.ind, population_size=5)

        fitness = lambda chromosome, island, x, y: (chromosome[0].value * x) + (chromosome[1].value * y)
        g_1 = Gene(name="first", value=1, gene_max=25, gene_min=1, randomise_function=random_int)
        g_2 = Gene(name="second", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_3 = Gene(name="third", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_4 = Gene(name="fourth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_5 = Gene(name="fifth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_6 = Gene(name="sixth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        gen = Chromosome([g_1, g_2, g_3, g_4, g_5, g_6])

        aliens = [Individual(fitness, name="AlsoAdam", chromosome=gen)]

        self.life.import_migrants(aliens)

        self.assertEquals(len(self.life.population), 5)

        fitness = lambda chromosome, island, x, y: (chromosome[0].value * x) + (chromosome[1].value * y)
        g_1 = Gene(name="first", value=5, gene_max=10, gene_min=1, randomise_function=random_int)
        g_2 = Gene(name="second", value=99, gene_max=100, gene_min=1, randomise_function=random_int)
        gen = Chromosome([g_1, g_2])

        aliens = [Individual(fitness, name="Eve", chromosome=gen)]

        self.life.import_migrants(aliens)

        self.assertEquals(len(self.life.population), 6)

    def test_evolve_generational(self):
        self.life.initialise(self.ind, population_size=5)
        self.life.evolve(crossover_params={'n_points' : 2})


from natural_selection.genetic_algorithms.operators.selection.elite_selection import *
from natural_selection.genetic_algorithms.operators.selection.survivor_selection import *
from natural_selection.genetic_algorithms.operators.selection.parent_selection import *


import unittest
from natural_selection.genetic_programs.functions import OperatorAdd, OperatorSub, OperatorMul, OperatorDiv, Operator
import natural_selection.genetic_programs.functions as op
from natural_selection.genetic_programs import Node

class TestNode(unittest.TestCase):

    def test_simple_add(self):
        n_3 = Node(label='3',arity=1,operator=None,terminal_value=3, is_terminal=True)
        n_39 = Node(arity=1,operator=None,terminal_value=39, is_terminal=True)
        n = Node(arity=2,operator=OperatorAdd(),children=[n_3,n_39])

        value = n()

        self.assertEquals(value, 42)
        self.assertEquals(str(n), "add(3, 39)")

    def test_simple_sub(self):
        n_42 = Node(terminal_value=42, is_terminal=True)
        n_2 = Node(terminal_value=2, is_terminal=True)
        n = Node(label='N',arity=2,operator=OperatorSub(),children=[n_42,n_2])

        value = n()

        self.assertEquals(value, 40)
        self.assertEquals(str(n), "N(42, 2)")

        n = Node(arity=2, operator=OperatorSub(), children=[n_2, n_42])

        value = n()

        self.assertEquals(value, -40)
        self.assertEquals(str(n), "sub(2, 42)")

    def test_simple_mul(self):
        n_1 = Node(terminal_value=2, is_terminal=True)
        n_2 = Node(terminal_value=2, is_terminal=True)
        n_3 = Node(terminal_value=2, is_terminal=True)

        n = Node(label='N',arity=2,operator=OperatorMul(),children=[n_1,n_2,n_3])

        value = n()

        self.assertEquals(value, 8)

    def test_simple_div(self):
        n_1 = Node(terminal_value=8, is_terminal=True)
        n_2 = Node(terminal_value=2, is_terminal=True)
        n_3 = Node(terminal_value=2, is_terminal=True)

        n = Node(label='N',arity=2,operator=OperatorDiv(),children=[n_1,n_2,n_3])

        value = n()

        self.assertEquals(value, 2)

    def test_custom_operator(self):
        import math
        def custom_func(args):
            return math.cos(args[0])

        operator_cos = Operator(operator_label='cos', function=custom_func)

        n_1 = Node(arity=1, is_terminal=True, terminal_value=0.5)
        n_2 = Node(arity=1, is_terminal=True, terminal_value=1.5)

        n_3 = Node('Nc',arity=1,operator=operator_cos, children=[n_1])
        n_4 = Node(arity=1,operator=operator_cos, children=[n_2])

        n_5 = Node(label='N',arity=1,operator=OperatorAdd(),children=[n_3,n_4])

        value = n_5()

        expected = math.cos(0.5) + math.cos(1.5)

        self.assertEquals(value, expected)
        self.assertEquals(str(n_5), "N(Nc(0.5), cos(1.5))")

    def test_custom_operator_kwargs(self):
        import math
        def custom_func(args):
            return math.cos(args[0])

        operator_cos = Operator(operator_label='cos', function=custom_func)

        n_1 = Node('x', is_terminal=True)
        n_2 = Node('z', is_terminal=True)

        n_3 = Node(label='cos',arity=1,operator=operator_cos,children=[n_1])
        n_4 = Node(operator=operator_cos, children=[n_2])

        n_5 = Node(operator=OperatorAdd(),children=[n_3,n_4])

        n_6 = Node(terminal_value=2, is_terminal=True)

        n_7 = Node(operator=OperatorAdd(), children=[n_5,n_6])

        value = n_7(x=0.5,z=1.5)

        expected = math.cos(0.5) + math.cos(1.5) + 2

        self.assertEquals(value, expected)

    def test_str(self):
        import math
        def custom_func(args):
            return math.cos(args[0])

        operator_cos = Operator(operator_label='cos', function=custom_func)

        n_1 = Node(label='x', arity=1, is_terminal=True)
        n_2 = Node(label='z', arity=1, is_terminal=True)

        n_3 = Node(operator=operator_cos, children=[n_1])
        n_4 = Node(operator=operator_cos, children=[n_2])

        n_5 = Node(label='add', arity=2, operator=OperatorAdd(),children=[n_3, n_4])

        n_6 = Node(label='2', arity=1, is_terminal=True, terminal_value=2)

        n_7 = Node(label='add', arity=2, operator=OperatorAdd(), children=[n_5, n_6])

        value = str(n_7)

        self.assertEquals(value, "add(add(cos(x), cos(z)), 2)")

        n_8 = Node(terminal_value=42, is_terminal=True)

        self.assertEquals(str(n_8), "42")

    def test_repr_simple(self):
        import math
        def custom_func(args):
            return math.cos(args[0])

        operator_cos = Operator(operator_label='cos', function=custom_func)

        n_1 = Node(label='x', arity=1, is_terminal=True)
        n_2 = Node(label='z', arity=1, is_terminal=True)

        n_3 = Node(operator=operator_cos, children=[n_1])
        n_4 = Node(operator=operator_cos, children=[n_2])

        n_5 = Node(label='add', arity=2, operator=OperatorAdd(),children=[n_3, n_4])

        n_6 = Node(label='2', arity=1, is_terminal=True, terminal_value=2)

        n_7 = Node(label='add', arity=2, operator=OperatorAdd(), children=[n_5, n_6])

        value = repr(n_7)

        self.assertNotEqual(value, "add(add(cos(x), cos(z)), 2)")
        self.assertEquals(value, 'add(2, add(cos(x), cos(z)))')

        n_8 = Node(terminal_value=42, is_terminal=True)

        self.assertEquals(repr(n_8), "42")

    def test_repr_advance(self):
        import math
        def custom_func(args):
            return math.cos(args[0])

        operator_cos = Operator(operator_label='cos', function=custom_func)

        n_1 = Node(label='z', arity=1, is_terminal=True)
        n_2 = Node(label='x', arity=1, is_terminal=True)

        n_3 = Node(operator=operator_cos, children=[n_1])
        n_4 = Node(operator=operator_cos, children=[n_2])

        n_5 = Node(label='add', arity=2, operator=OperatorAdd(),children=[n_3, n_4])

        n_6 = Node(label='2', arity=1, is_terminal=True, terminal_value=2)

        n_7 = Node(label='add', arity=2, operator=OperatorAdd(), children=[n_5, n_6])

        value_1 = repr(n_7)

        n_1 = Node(label='x', arity=1, is_terminal=True)
        n_2 = Node(label='z', arity=1, is_terminal=True)

        n_3 = Node(operator=operator_cos, children=[n_1])
        n_4 = Node(operator=operator_cos, children=[n_2])

        n_5 = Node(label='add', arity=2, operator=OperatorAdd(), children=[n_3, n_4])

        n_6 = Node(label='2', arity=1, is_terminal=True, terminal_value=2)

        n_7 = Node(label='add', arity=2, operator=OperatorAdd(), children=[n_5, n_6])

        value_2 = repr(n_7)

        self.assertEquals(value_1, value_2)

        n_1 = Node(label='x', arity=1, is_terminal=True)
        n_2 = Node(arity=1, is_terminal=True, terminal_value=3)

        n_3 = Node(operator=operator_cos, children=[n_1])

        n_5 = Node(label='add', arity=2, operator=OperatorAdd(), children=[n_2, n_3])

        n_6 = Node(label='2', arity=1, is_terminal=True, terminal_value=2)

        n_7 = Node(label='add', arity=2, operator=OperatorAdd(), children=[n_5, n_6])

        value_1 = repr(n_7)
        str_1 = str(n_7)

        n_1 = Node(label='x', arity=1, is_terminal=True)
        n_2 = Node(arity=1, is_terminal=True, terminal_value=3)

        n_3 = Node(operator=operator_cos, children=[n_1])

        n_5 = Node(label='add', arity=2, operator=OperatorAdd(), children=[n_3, n_2])

        n_6 = Node(label='2', arity=1, is_terminal=True, terminal_value=2)

        n_7 = Node(label='add', arity=2, operator=OperatorAdd(), children=[n_6, n_5])

        value_2 = repr(n_7)
        str_2 = str(n_7)

        self.assertEquals(value_1, value_2)
        self.assertNotEqual(str_1, str_2)

    def test_return_node(self):
        x = Node(label='x', is_terminal=True)
        n = Node(operator=op.OperatorReturn(), children=[x])

        value = n(x=42)

        self.assertEquals(value, 42)
        self.assertEquals(str(n), "return(x)")

    def test_depth(self):
        n_1 = Node(terminal_value=2, is_terminal=True)
        n_2 = Node(terminal_value=2, is_terminal=True)

        n_mul = Node(label='N',arity=2,operator=OperatorMul(),children=[n_1,n_2])

        n_3 = Node(terminal_value=5, is_terminal=True)

        n_add = Node(operator=OperatorAdd(), children=[n_mul,n_3])

        value = n_add.depth()

        self.assertEquals(value, 3)

        n_x = Node(label='x', is_terminal=True)
        n_min = Node(operator=OperatorSub(), children=[n_add, n_x])

        value = n_min.depth()

        self.assertEquals(value, 4)

        value = n_1.depth()

        self.assertEquals(value, 1)

        value = n_mul.depth()

        self.assertEquals(value, 2)


from numpy import random

def random_int(gene):
    """
    Random integer from range.

    Args:
        gene (Gene): A gene with a set `gene_min` and `gene_max`.

    Returns:
        int: Random number.
    """
    return random.randint(low=gene.gene_min, high=gene.gene_max)

def random_int_step(gene):
    """
    Random integer step from a range. Stepping can be  solely to the right (increment), left (decrement), or in both.
    This depends on `step_lower_bound` and `step_upper_bound` values of the gene. For left steps only, both need to be negative.
    For right only, both should be positive values.

    stepped_value = value + random(int, int)

    Stepped values are still bound to the `gene_min` and `gene_max` constraints and will continue setting until condition is matched.

    Args:
         gene (Gene): A gene with a set `step_lower_bound` and `step_upper_bound`.

    Returns:
        int: Stepped value.
    """
    stepped_value = gene.value + random.randint(low=gene.step_lower_bound, high=gene.step_upper_bound)

    while stepped_value > gene.gene_max or stepped_value < gene.gene_min:
        stepped_value = gene.value + random.randint(low=gene.step_lower_bound, high=gene.step_upper_bound)

    return stepped_value

def random_gaussian(gene):
    """
    Random float from gaussian dist.

    Args:
        gene (Gene): A gene with a set `mu` and `sig`.

    Returns:
        float: Random number.
    """
    stepped_value = random.normal(loc=gene.mu, scale=gene.sig)

    while stepped_value > gene.gene_max or stepped_value < gene.gene_min:
        stepped_value = random.normal(loc=gene.mu, scale=gene.sig)

    return stepped_value

def random_uniform(gene):
    """
    Random float from uniform dist.

    Args:
        gene (Gene): A gene with a set `gene_min` and `gene_max`.

    Returns:
        float: Random number.
    """
    return random.uniform(low=gene.gene_min, high=gene.gene_max)

def random_gaussian_step(gene):
    """
    Random gaussian step. Stepping can be  solely to the right (increment), left (decrement), or in both.
    This depends on `mu` and `sig` values of the gene. For left steps only, both need to be negative.
    For right only, both should be positive values. These need to be considered in advance.

    stepped_value = value + gaussian(mu, sig)

    Stepped values are still bound to the `gene_min` and `gene_max` constraints and will continue setting until condition is matched.

    Args:
         gene (Gene): A gene with a set `mu` and `sig`.

    Returns:
        float: Stepped value.
    """
    stepped_value = gene.value + random.normal(loc=gene.mu, scale=gene.sig)

    while stepped_value > gene.gene_max or stepped_value < gene.gene_min:
        stepped_value = gene.value + random.normal(loc=gene.mu, scale=gene.sig)

    return stepped_value

def random_uniform_step(gene):
    """
    Random uniform step from a range. Stepping can be  solely to the right (increment), left (decrement), or in both.
    This depends on `step_lower_bound` and `step_upper_bound` values of the gene. For left steps only, both need to be negative.
    For right only, both should be positive values.

    stepped_value = value + uniform(l, u)

    Stepped values are still bound to the `gene_min` and `gene_max` constraints and will continue setting until condition is matched.

    Args:
         gene (Gene): A gene with a set `step_lower_bound` and `step_upper_bound`.

    Returns:
        float: Stepped value.
    """
    stepped_value = gene.value + random.uniform(low=gene.step_lower_bound, high=gene.step_upper_bound)

    while stepped_value > gene.gene_max or stepped_value < gene.gene_min:
        stepped_value = gene.value + random.uniform(low=gene.step_lower_bound, high=gene.step_upper_bound)

    return stepped_value

def random_choice(gene):
    """
    Randomly select a object, such as strings, from a list. Gene must have defined `choices` list.

    Args:
        gene (Gene): A gene with a set `choices` list.

    Returns:
        object: Selected choice.
    """
    if not 'choices' in gene.__dict__:
        raise KeyError("'choices' not defined in this gene, please include a list values!")
    return random.choice(gene.choices)


# -*- coding: utf-8 -*-
"""Basic classes for running Genetic Algorithms.
"""
__author__ = "Justin Hocking"
__copyright__ = "Copyright 2021, Zipfian Science"
__credits__ = []
__license__ = ""
__version__ = "0.1.0"
__maintainer__ = "Justin Hocking"
__email__ = "justin.hocking@zipfian.science"
__status__ = "Development"

import random
import uuid
from typing import Callable, Any, Iterable


from natural_selection.genetic_algorithms.operators.initialisation import initialise_population_random
from natural_selection.genetic_algorithms.operators.selection import selection_elites_top_n, selection_parents_two
from natural_selection.genetic_algorithms.operators.crossover import crossover_two_uniform
from natural_selection.genetic_algorithms.operators.mutation import mutation_randomize
from natural_selection.genetic_algorithms.utils.probability_functions import crossover_prob_function_classic, mutation_prob_function_classic
from natural_selection.genetic_algorithms.utils import clone_classic, GeneticAlgorithmError

class Gene:
    """
    A simple class to encapsulate a simple gene.

    Args:
        name (str): Gene name. The gene name also acts as a compatibility reference.
        value (Any): The value, could be any type.
        randomise_function (Callable): A function to randomise the gene, taking the gene (`self`) as input with signature ``func(self)``.
        gene_max (Any, numeric type): Max value for random number generator (default = None).
        gene_min (Any, numeric type): Min value for random number generator (default = None).
        mu (Any, numeric type): Mean value of distribution to sample from (default = 0).
        sig (Any, numeric type): Std. Dev. value of distribution to sample from (default = 1).
        step_lower_bound (Any, numeric type): For uniform stepping functions, defines lower bound of range (default = -1.0).
        step_upper_bound (Any, numeric type): For uniform stepping functions, defines upper bound of range (default = 1.0).
        choices (Iterable): List of choices, categorical or not, to randomly choose from (default = None).
        gene_properties (dict): For custom random functions, extra params may be given (default = None).

    """

    def __init__(self,
                 name : str,
                 value : Any,
                 randomise_function: Callable,
                 gene_max : Any = None,
                 gene_min : Any = None,
                 mu : Any = 0,
                 sig: Any = 1,
                 step_lower_bound : Any = -1.0,
                 step_upper_bound: Any = 1.0,
                 choices : Iterable  = None,
                 gene_properties : dict = None):
        self.name = name
        self.value = value
        self.gene_max = gene_max
        self.gene_min = gene_min
        self.mu = mu
        self.sig = sig
        self.step_lower_bound = step_lower_bound
        self.step_upper_bound = step_upper_bound
        self.choices = choices
        self.randomise_function = randomise_function
        self.__gene_properties = gene_properties
        if gene_properties:
            for k, v in gene_properties.items():
                self.__dict__.update({k:v})

    def randomise(self):
        """
        Sets a random value gene with randomised value.
        """
        self.value = self.randomise_function(gene=self)

    def randomise_new(self):
        """
        Return a new gene with randomised value.

        Returns:
            Gene: Newly created gene.
        """
        return Gene(
            name=self.name,
            value=self.randomise_function(gene=self),
            randomise_function=self.randomise_function,
            gene_max=self.gene_max,
            gene_min=self.gene_min,
            mu=self.mu,
            sig=self.sig,
            choices=self.choices,
            gene_properties=self.__gene_properties
        )

    def add_new_property(self, key : str, value : Any):
        if self.__gene_properties:
            self.__gene_properties.update({key: value})
        else:
            self.__gene_properties = {key: value}
        self.__dict__.update({key: value})

    def __str__(self) -> str:
        return f'Gene({self.name}:{self.value})'

    def __repr__(self):
        start_str = f'Gene({self.name}:{self.value}:{self.gene_max}:{self.gene_min}:{self.mu}:{self.sig}:{self.step_lower_bound}:{self.step_upper_bound}'
        if self.choices:
            for c in self.choices:
                start_str = f'{start_str}:{str(c)}'
        start_str = f'{start_str}:{str(self.randomise_function)}'
        start_str = f'{start_str}:{str(self.__gene_properties)})'
        return  start_str


class Chromosome:
    """
    A class that encapsulates an ordered sequence of Gene objects.

    Note:
        gene_verify_func should take the gene, index of gene, and chromosome (`self`) as parameters.

    Args:
        genes (list): list of initialised Gene objects.
        gene_verify_func (Callable): A function to verify gene compatibility `func(gene,loc,chromosome)` (default = None).
    """

    def __default_gene_verify_func(self, gene, loc, chromosome):
        """
        Needed to make objects pickle-able.
        Args:
            gene (Gene): Gene being inserted.
            loc (int): Index of gene.
            chromosome (Chromosome): The given "self".

        Returns:
            bool: Whether gene insertion is allowed or not.
        """
        return True

    def __init__(self, genes: list = None, gene_verify_func : Callable = None):
        if genes:
            self.genes = genes
        else:
            self.genes = list()

        if gene_verify_func:
            self.gene_verify_func = gene_verify_func
        else:
            self.gene_verify_func = self.__default_gene_verify_func

    def append(self, gene: Gene):
        """
        Simple appending of Gene type objects.

        Args:
            gene (Gene): Gene
        """
        assert isinstance(gene, Gene), 'Must be Gene type!'
        if not self.gene_verify_func(gene=gene,loc=-1,chromosome=self):
            raise GeneticAlgorithmError(message="Added gene did not pass compatibility tests!")
        self.genes.append(gene)

    def __setitem__(self, index, gene):
        if isinstance(index, slice):
            assert index.start < len(self.genes), 'Index Out of bounds!'
        else:
            assert isinstance(gene, Gene), 'Must be Gene type!'
            assert index < len(self.genes), 'Index Out of bounds!'

        if not self.gene_verify_func(gene=gene, loc=index, chromosome=self):
            raise GeneticAlgorithmError("Index set gene did not pass compatibility tests!")
        self.genes[index] = gene

    def __getitem__(self, index) -> Gene:
        if isinstance(index, slice):
            assert index.start < len(self.genes), 'Index Out of bounds!'
        else:
            assert index < len(self.genes), 'Index Out of bounds!'

        return self.genes[index]

    def __iter__(self):
        self.__n = 0
        return self

    def __next__(self):
        if self.__n < len(self.genes):
            gene = self.genes[self.__n]
            self.__n += 1
            return gene
        else:
            raise StopIteration

    def __len__(self):
        return len(self.genes)

    def __str__(self) -> str:
        start_str = '-'.join([str(t) for t in self.genes])
        return f'Chromosome({start_str})'

    def __repr__(self):
        start_str = '-'.join([repr(t) for t in self.genes])
        start_str = f'Chromosome({start_str})'
        return start_str


class Individual:
    """
    A class that encapsulates a single individual, with genetic code and a fitness evaluation function.

    Args:
        fitness_function (Callable): Function with ``func(Chromosome, island, **params)`` signature.
        name (str): Name for keeping track of lineage (default = None).
        chromosome (Chromosome): A Chromosome object, initialised (default = None).
        species_type (str) : A unique string to identify the species type, for preventing cross polluting (default = None).

    Attributes:
        fitness (Numeric): The fitness score after evaluation.
        age (int): How many generations was the individual alive.
        genetic_code (str): String representation of Chromosome.
        history (list): List of dicts of every evaluation.
        parents (list): List of strings of parent names.
    """

    def __init__(self, fitness_function : Callable, name : str = None, chromosome: Chromosome = None, species_type : str = None):
        if name is None:
            self.name = str(uuid.uuid4())
        else:
            self.name = name
        if chromosome is None:
            self.chromosome = Chromosome()
        else:
            self.chromosome = chromosome
        self.fitness_function = fitness_function
        self.fitness = None
        self.age = 0
        self.genetic_code = None
        self.history = list()
        self.parents = list()
        if species_type:
            self.species_type = species_type
        else:
            self.species_type = "DEFAULT_SPECIES"

    def register_parent_names(self, parents : list, reset_parent_name_list : bool = True):
        """
        In keeping lineage of family lines, the names of parents are kept track of.

        Args:
            parents (list): A list of Individuals of the parents.
        """
        if reset_parent_name_list:
            self.parents = list()
        for parent in parents:
            self.parents.append(parent.name)

    def reset_name(self, name : str = None):
        """
        A function to reset the name of an individual, helping to keep linage of families.

        Args:
            name (str): Name (default = None).
        """
        if name is None:
            self.name = str(uuid.uuid4())
        else:
            self.name = name

    def birthday(self, add : int = 1):
        """
        Add to the age. This is for keeping track of how many generations an individual has "lived" through.

        Args:
            add (int): Amount to age.
        """
        self.age += add

    def reset_fitness(self, fitness : Any = None, reset_genetic_code : bool = True):
        """
        Reset (or set) the fitness of the individual.

        Args:
            fitness (Any): New fitness value (default = None).
            reset_genetic_code (bool): Whether to reset the genetic code. (default = True)
        """
        self.fitness = fitness
        if reset_genetic_code:
            self.genetic_code = None

    def add_gene(self, gene : Gene):
        """
        Appends a gene to the chromosome.

        Args:
            gene (Gene): Gene to add.
        """
        self.chromosome.append(gene)

    def evaluate(self, params : dict, island=None) -> Any:
        """
        Run the fitness function with the given params.

        Args:
            params (dict): Named dict of eval params.
            island (Island): Pass the Island for advanced fitness functions based on Island properties and populations.

        Returns:
            numeric: Fitness value.
        """
        self.fitness = self.fitness_function(chromosome=self.chromosome, island=island, **params)

        stamp = { "name": self.name,
                  "age": self.age,
                  "fitness": self.fitness,
                  "chromosome": str(self.chromosome),
                  "parents": self.parents,
         }

        if island:
            stamp["island_generation" ] = island.generation_count

        self.history.append(stamp)
        return self.fitness

    def unique_genetic_code(self) -> str:
        """
        Gets the unique genetic code, generating if it is undefined.

        Returns:
            str: String name of Chromosome.
        """
        if self.genetic_code is None:
            self.genetic_code = repr(self.chromosome)
        return self.genetic_code

    def __str__(self) -> str:
        return f'Individual({self.name}:{self.fitness})'

    def __repr__(self) -> str:
        genetic_code = self.unique_genetic_code()
        return f'Individual({self.name}:{self.fitness}:{self.age}:{self.species_type}:{genetic_code})'

    def __eq__(self, other):
        if isinstance(other, Individual):
            return self.unique_genetic_code() == other.unique_genetic_code()
        else:
            raise GeneticAlgorithmError(message=f'Can not compare {type(other)}')

    def __ne__(self, other):
        if isinstance(other, Individual):
            return self.unique_genetic_code() != other.unique_genetic_code()
        else:
            raise GeneticAlgorithmError(message=f'Can not compare {type(other)}')

    def __lt__(self, other):
        if isinstance(other, Individual):
            return self.fitness < other.fitness
        elif isinstance(other, int):
            return self.fitness < other
        elif isinstance(other, float):
            return self.fitness < other
        else:
            raise GeneticAlgorithmError(message=f'Can not compare {type(other)}')

    def __le__(self, other):
        if isinstance(other, Individual):
            return self.fitness <= other.fitness
        elif isinstance(other, int):
            return self.fitness <= other
        elif isinstance(other, float):
            return self.fitness <= other
        else:
            raise GeneticAlgorithmError(message=f'Can not compare {type(other)}')

    def __gt__(self, other):
        if isinstance(other, Individual):
            return self.fitness > other.fitness
        elif isinstance(other, int):
            return self.fitness > other
        elif isinstance(other, float):
            return self.fitness > other
        else:
            raise GeneticAlgorithmError(message=f'Can not compare {type(other)}')

    def __ge__(self, other):
        if isinstance(other, Individual):
            return self.fitness >= other.fitness
        elif isinstance(other, int):
            return self.fitness >= other
        elif isinstance(other, float):
            return self.fitness >= other
        else:
            raise GeneticAlgorithmError(message=f'Can not compare {type(other)}')

class Island:
    """
    A simple Island to perform a Genetic Algorithm. By default the selection, mutation, crossover, and probability functions
    default to the classic functions.

    Args:
        function_params (dict): The parameters for the fitness function.
        elite_selection (Callable): Function for selecting individuals for crossover and mutation (default = None).
        parent_selection (Callable): Function for selecting parents for crossover (default = None).
        crossover_function (Callable): Function for crossover (default = None).
        mutation_function (Callable): Function for mutation (default = None).
        crossover_prob_function (Callable): Random probability function for crossover (default = None).
        mutation_prob_function (Callable): Random probability function for mutation (default = None).
        clone_function (Callable): Function for cloning (default = None).
        random_seed (int): Random seed for random and Numpy generators, set to None for no seed (default = 42).
        verbose (bool): Print all information (default = None).
        logging_function (Callable): Function for custom message logging, such as server logging (default = None).
        force_genetic_diversity (bool): Only add new offspring to the population if they have a unique chromosome (default = True).

    Attributes:
        unique_genome (list): List of unique chromosomes.
        generation_info (list): List of dicts detailing info for every generation.
        population (list): The full population of members.
        elites (list): All elites selected during the run.
        mutants (list): All mutants created during the run.
        children (list): All children created during the run.
        generation_count (int): The current generation number.
    """

    def __init__(self, function_params : dict,
                 elite_selection : Callable = selection_elites_top_n,
                 parent_selection : Callable = selection_parents_two,
                 crossover_function : Callable = crossover_two_uniform,
                 mutation_function : Callable = mutation_randomize,
                 crossover_prob_function : Callable = crossover_prob_function_classic,
                 mutation_prob_function : Callable = mutation_prob_function_classic,
                 clone_function : Callable = clone_classic,
                 random_seed: int = 42,
                 verbose : bool = True,
                 logging_function : Callable = None,
                 force_genetic_diversity : bool = True):
        self.function_params = function_params
        self.unique_genome = list()
        self.generation_info = list()
        self.population = list()


        self.elite_selection = elite_selection
        self.parent_selection = parent_selection
        self.crossover = crossover_function
        self.mutation = mutation_function
        self.crossover_prob = crossover_prob_function
        self.mutation_prob = mutation_prob_function
        self.clone = clone_function

        self.logging_function = logging_function
        self.force_genetic_diversity = force_genetic_diversity

        self.verbose = verbose
        self.random_seed = random_seed
        self.elites = list()
        self.mutants = list()
        self.children = list()
        self.species_type = "DEFAULT_SPECIES"
        self.generation_count = 0

        # Set python random seed, as well as Numpy seed.
        if random_seed:
            random.seed(random_seed)
            from numpy.random import seed as np_seed
            np_seed(random_seed)

    def create_gene(self,
                 name : str,
                 value : Any,
                 randomise_function: Callable,
                 gene_max : Any = None,
                 gene_min : Any = None,
                 mu : Any = None,
                 sig: Any = None,
                 choices : Iterable  = None,
                 gene_properties : dict = None):
        """
        Wrapping function to create a new Gene. Useful when writing new initialisation functions. See Gene class.

        Args:
            name (str): Gene name. The gene name also acts as a compatibility reference.
            value (Any): The value, could be any type.
            randomise_function (Callable): A function to randomise the gene, taking the min and max as input with signature ``func(self)``.
            gene_max (Any, numeric type): Max value for random number generator (default = None).
            gene_min (Any, numeric type): Min value for random number generator (default = None).
            mu (Any, numeric type): Mean value of distribution to sample from (default = None).
            sig (Any, numeric type): Std. Dev. value of distribution to sample from (default = None).
            choices (Iterable): List of choices, categorical or not, to randomly choose from (default = None).
            gene_properties (dict): For custom random functions, extra params may be given (default = None).
        Returns:
            gene: A new Gene object.
        """
        return Gene(name, value, randomise_function, gene_max, gene_min, mu, sig, choices, gene_properties)

    def create_chromosome(self,
                          genes: list = None,
                          gene_verify_func : Callable = None):
        """
        Wrapping function to create a new Chromosome. Useful when writing new initialisation functions. See Chromosome class.

        Args:
            genes (list): list of initialised Gene objects.
            gene_verify_func (Callable): A function to verify gene compatibility `func(gene,loc,chromosome)` (default = None).

        Returns:
            chromosome: A new Chromosome.
        """
        return Chromosome(genes, gene_verify_func)

    def create_individual(self,
                          fitness_function : Callable,
                          name : str = None,
                          chromosome: Chromosome = None,
                          species_type : str = None,
                          add_to_population : bool = False):
        """
        Wrapping function to create a new Individual. Useful when writing new initialisation functions. See Individual class.

        Args:
            fitness_function (Callable): Function with ``func(Chromosome, island, **params)`` signature
            name (str): Name for keeping track of lineage (default = None).
            chromosome (Chromosome): A Chromosome object, initialised (default = None).
            species_type (str) : A unique string to identify the species type, for preventing cross polluting (default = None).
            add_to_population (bool): Add this new individual to the population (default = False).

        Returns:
            individual: A new Individual.
        """
        ind = Individual(fitness_function, name, chromosome, species_type)
        if add_to_population:
            self.population.append(ind)
        return ind

    def initialise(self, adam : Individual,
               population_size: int = 8,
               initialisation_function : Callable = initialise_population_random,
               initialisation_params : dict = {},
               evaluate_population : bool = True
               ):
        """
        Starts the population by taking an initial individual as template and creating new ones from it.

        Args:
            adam (Individual): Individual to clone from.
            population_size (int): Size of population.
            initialisation_function (Callable): A function for randomly creating new individuals from the given adam.
            initialisation_params (dict): Custom params for custom initialisation functions.
            evaluate_population (bool): Evaluate the newly created population (default = True).
        """
        self.species_type = adam.species_type

        self.initialise = initialisation_function

        self.population = self.initialise(adam=adam, n=population_size, island=self, **initialisation_params)

        if evaluate_population:
            for popitem in self.population:
                popitem.evaluate(self.function_params, island=self)
                self.unique_genome.append(popitem.unique_genetic_code())

    def import_migrants(self, migrants : list,
                        reset_fitness : bool = False,
                        species_check : bool = True,
                        force_genetic_diversity : bool = True):
        """
        Imports a list of individuals, with option to re-evaluate them.
        Skips the individual if the genetic code is already in the population.

        Args:
            migrants (list): List of Individuals.
            reset_fitness (bool): Reset the fitness of new members and evaluate them (default = False).
            species_check (bool): Safely check that imported members are compatible with population  (default = True).
            force_genetic_diversity (bool): Only add migrants to the population if they have a unique chromosome (default = True).
        """
        for i in migrants:
            if species_check:
                if i.species_type != self.species_type:
                    continue
            if force_genetic_diversity:
                if not i.unique_genetic_code() in self.unique_genome:
                    if i.fitness is None or reset_fitness:
                        i.evaluate(self.function_params, island=self)
                    self.population.append(i)
                    self.unique_genome.append(i.unique_genetic_code())
            else:
                if i.fitness is None or reset_fitness:
                    i.evaluate(self.function_params)
                self.population.append(i)
                self.unique_genome.append(i.unique_genetic_code())


    def evolve(self, starting_generation : int = 0,
                            n_generations : int = 5,
                            crossover_probability : float = 0.5,
                            mutation_probability : float = 0.5,
                            crossover_params : dict = None,
                            mutation_params : dict = None,
                            elite_selection_params : dict = None,
                            parent_selection_params: dict = None,
                            survivor_selection_params: dict = None,
                            criterion_function : Callable = None,
                            criterion_params : dict = None) -> Individual:
        """
        Starts the evolutionary run.

        Args:
            starting_generation (int): Starting generation.
            n_generations (int): Number of generations to run.
            crossover_probability (float): Initial crossover probability.
            mutation_probability (float): Initial mutation probability.
            crossover_params (dict): Dict of params for custom crossover function (default = None).
            mutation_params (dict): Dict of params for custom mutation function (default = None).
            selection_params (dict): Dict of params for custom selection function (default = None).
            criterion_function (Callable): A function to evaluate if the desired criterion has been met (default = None).
            criterion_params (dict): Function parameters for criterion (default = None).

        Returns:
            Individual: The fittest Individual found.
        """

        if crossover_params:
            _crossover_params = crossover_params
        else:
            _crossover_params = {'prob':0.5}

        if mutation_params:
            _mutation_params = mutation_params
        else:
            _mutation_params = {'prob':0.2}

        if elite_selection_params:
            _elite_selection_params = elite_selection_params
        else:
            _elite_selection_params = {'n':5, 'desc':True}

        if parent_selection_params:
            _parent_selection_params = parent_selection_params
        else:
            _parent_selection_params = {}

        if survivor_selection_params:
            _survivor_selection_params = survivor_selection_params
        else:
            _survivor_selection_params = {}

        def _default_criterion_function(island):
            return island.generation_count < starting_generation + n_generations - 1

        if criterion_function:
            g_func = criterion_function
        else:
            g_func = _default_criterion_function
            criterion_params = {}

        g = starting_generation
        while g_func(island=self, **criterion_params):
            self._verbose_logging('Generation {} started'.format(g))

            self.__evolutionary_engine(g=g,
                                       elite_selection_params=_elite_selection_params,
                                       parent_selection_params=_parent_selection_params,
                                       survivor_selection_params=_survivor_selection_params,
                                       crossover_probability=crossover_probability,
                                       mutation_probability=mutation_probability,
                                       crossover_params=_crossover_params,
                                       mutation_params=_mutation_params)
            g += 1

        best_ind = selection_elites_top_n(island=self, individuals=self.population, n=1)[0]

        self._verbose_logging("Best individual is {0}, {1}".format(best_ind.name, best_ind.fitness))

        return best_ind

    def __evolutionary_engine(self,
                              g,
                              elite_selection_params,
                              parent_selection_params,
                              survivor_selection_params,
                              crossover_probability,
                              mutation_probability,
                              crossover_params,
                              mutation_params):

        elites = self.elite_selection(island=self,
                                      individuals=self.clone(individuals=self.population, island=self),
                                      **elite_selection_params)

        self.elites.append({'generation' : g, 'elites' : elites})

        # Children are strictly copies or new objects seeing as the have a lineage and parents
        generation_children = list()
        for parents in self.parent_selection(individuals=elites, island=self, **parent_selection_params):
            if self.crossover_prob(crossover_probability=crossover_probability, island=self):

                self._verbose_logging('Crossover: {0}'.format([str(p) for p in parents]))

                children = self.crossover(island=self,
                                          individuals=self.clone(individuals=parents, island=self),
                                          **crossover_params)

                for child in children:
                    child.reset_name()
                    child.register_parent_names(parents)
                    child.age = 0
                    child.reset_fitness()

                generation_children.extend(children)

        self.children.append({'generation' : g, 'children' : generation_children})

        # Mutants are not strictly copied but rather only modified seeing as the are part of the children list
        generation_mutants = list()
        for mutant in generation_children:
            if self.mutation_prob(mutation_probability=mutation_probability, island=self):
                mutated = self.mutation(island=self, individual=mutant, **mutation_params)

                self._verbose_logging('Mutating: {}'.format(mutant.name))
                mutated.reset_fitness()
                generation_mutants.append(mutated)

        self.mutants.append({'generation': g, 'mutants': generation_mutants})

        untested_individuals = [ind for ind in generation_children if ind.fitness is None]

        for new_untested_individual in untested_individuals:
            new_untested_individual.evaluate(island=self, params=self.function_params)
            if self.force_genetic_diversity:
                # If we want a diverse gene pool, this must be true
                if not new_untested_individual.unique_genetic_code() in self.unique_genome:
                    self.population.append(new_untested_individual)
                    self.unique_genome.append(new_untested_individual.unique_genetic_code())
            else:
                # Else, add it effectively allowing "twins" to exist
                self.population.append(new_untested_individual)
                self.unique_genome.append(new_untested_individual.unique_genetic_code())


        population_fitnesses = [ind.fitness for ind in self.population]
        elite_fitnesses = [ind.fitness for ind in elites]

        population_length = len(self.population)
        population_mean = sum(population_fitnesses) / population_length
        population_sum2 = sum(x * x for x in population_fitnesses)
        population_std = abs(population_sum2 / population_length - population_mean ** 2) ** 0.5

        self.generation_info.append(
            {
                'stat': 'population',
                'generation': g,
                'pop_len': population_length,
                'fitness_mean': population_mean,
                'fitness_std': population_std,
                'fitness_min': min(population_fitnesses),
                'fitness_max': max(population_fitnesses),
            }
        )

        self._verbose_logging(self.generation_info[-1])

        elite_length = len(elites)
        elite_mean = sum(elite_fitnesses) / elite_length
        elite_sum2 = sum(x * x for x in elite_fitnesses)
        elite_std = abs(elite_sum2 / elite_length - elite_mean ** 2) ** 0.5

        self.generation_info.append(
            {
                'stat': 'elites',
                'generation': g,
                'pop_len': elite_length,
                'fitness_mean': elite_mean,
                'fitness_std': elite_std,
                'fitness_min': min(elite_fitnesses),
                'fitness_max': max(elite_fitnesses),
            }
        )

        self._verbose_logging(self.generation_info[-1])

        for i in self.population:
            i.birthday()

        self.generation_count = g

    def _verbose_logging(self, message):
        if self.verbose:
            print(message)
        if self.logging_function:
            self.logging_function(message)



def initialise_population_random(adam, n : int, island=None):
    """
    Classic random initialisation function to create a pool of `n` individuals from a starting Individual `adam`.

    Args:
        adam (Individual): An individual already initialised with a chromosome.
        n (int): Population size.
        island (Island): Needed to wrap to `create_chromosome` and `create_individual` methods.

    Returns:
        list: Population members.
    """
    population = list()

    for i in range(n - 1):
        chromosome = island.create_chromosome([x.randomise_new() for x in adam.chromosome])
        eve = island.create_individual(adam.fitness_function, chromosome=chromosome)
        population.append(eve)

    population.append(adam)

    return population


import unittest
import random
from natural_selection.genetic_algorithms import Gene, Chromosome
from natural_selection.genetic_algorithms.utils.random_functions import random_int

class TestGenome(unittest.TestCase):

    def test_init(self):
        g = Gene(name="test", value=3, gene_max=10, gene_min=1, randomise_function=random_int)
        gen = Chromosome([g])

        self.assertGreater(len(gen.genes), 0)

    def test_append(self):
        g = Gene(name="test", value=3, gene_max=10, gene_min=1, randomise_function=random_int)
        gen = Chromosome()
        gen.append(g)

        with self.assertRaises(AssertionError) as ae:
            gen.append(1)

    def test_indexing(self):
        g = Gene(name="test", value=3, gene_max=10, gene_min=1, randomise_function=random_int)
        gen = Chromosome([g])

        self.assertEquals(gen[0], g)

        with self.assertRaises(AssertionError) as ae:
            gen[2]

        with self.assertRaises(AssertionError) as ae:
            gen[0] = 1

        gen[0] = Gene("test_other", 3, 10, 1, random.randint)

        self.assertNotEqual(gen[0], g)

    def test_len_and_str_and_iter(self):
        g_1 = Gene(name="test", value=3, gene_max=10, gene_min=1, randomise_function=random_int)
        g_2 = Gene(name="test_other", value=4, gene_max=10, gene_min=1, randomise_function=random_int)
        gen = Chromosome([g_1, g_2])

        self.assertEquals(len(gen), 2)
        self.assertEquals(str(gen), 'Chromosome(Gene(test:3)-Gene(test_other:4))')

        for g in gen:
            self.assertIsInstance(g, Gene)

        r = repr(gen)
        s = str(gen)

        self.assertNotEqual(r,s)


import random

def mutation_prob_function_classic(mutation_probability : float = 0.2, island=None) -> float:
    """
    Classic mutation probability function that evaluates whether a random float is less than the mutation probability.

    Args:
        mutation_probability (float): Probability of mutation.
        island (Island): The Island calling the method (optional, default = None).

    Returns:
        bool: Returns whether to perform mutation.
    """
    if random.random() < mutation_probability:
        return True
    else:
        return False


def crossover_prob_function_classic(crossover_probability : float = 0.5, island=None):
    """
    Classic crossover probability function that evaluates whether a random float is less than the crossover probability.

    Args:
        crossover_probability (float): Probability of crossover.
        island (Island): The Island calling the method (optional, default = None).

    Returns:
        bool: Returns whether to perform crossover.
    """
    if random.random() < crossover_probability:
        return True
    else:
        return False